#include "stdafx.h"
#include "assembler.h"
#include "asmreader.h"

#include <stdio.h>
#include <stdlib.h>

// Register access
#define REG_B 0x00
#define REG_C 0x01
#define REG_D 0x02
#define REG_E 0x03
#define REG_H 0x04
#define REG_L 0x05
#define REG_NULL 0x06
#define REG_A 0x07

Assembler::Assembler():
m_MemoryCounter(0)
{
    memset(m_Memory, 0, MEMORY_SIZE);
}

Assembler::~Assembler()
{
    if(m_pReader)
    {
        delete m_pReader;
    }

}

void Assembler::Init(const char* pFilename)
{
    m_pReader = new AsmReader(pFilename);
}

int Assembler::Assemble()
{
    if(!m_pReader)
        return 1;

    bool bDone = false;
    for (int i = 0; !bDone; i++)
    {
        AsmArgs args;
        m_pReader->GetLine(&args);
       
        if(args.argc < 1)
        {
            bDone = true;
        }

       int res = HandleLine(&args);
        if(res != 0)
        {
            printf("Error! on line %d\n", m_pReader->GetCurrentLineNumber());
            return 1;
        }
    }

    return 0;
}

int Assembler::HandleLine(AsmArgs* pArgs)
{
    // Print out the stats on the line
    /*
    for(int i = 0; i < args.argc; i++)
    {
    printf("%s ", args.argv[i]);
    }
    printf(" n=%d\n", args.argc);
    */
    if(pArgs->argc < 1)
    {
        return 0;
    }

    // Find out which instruction this is.
    // MOV
    if (strcmpi("MOV", pArgs->argv[0]) == 0)
    {
        //printf("MOV\n");
        return HandleOpcodeMOV(pArgs);
    }
    else
    // MVI
    if (strcmpi("MVI", pArgs->argv[0]) == 0)
    {
        //printf("MVI\n");
        return HandleOpcodeMVI(pArgs);
    }
    else
    // ADD
    if (strcmpi("ADD", pArgs->argv[0]) == 0)
    {
        //printf("ADD\n");
        return HandleOpcodeADD(pArgs);
    }
    else
    // CMP
    if (strcmpi("CMP", pArgs->argv[0]) == 0)
    {
        //printf("CMP\n");
        return HandleOpcodeCMP(pArgs);
    }
    else
    // RET
    if (strcmpi("RET", pArgs->argv[0]) == 0)
    {
        //printf("RET\n");
        return HandleOpcodeRET(pArgs);
    }
    else
    {
        printf("Undefined Instruction\n");
        return 1;
    }

    return 0;
}


int Assembler::HandleOpcodeMOV(AsmArgs* pArgs)
{
    if(pArgs->argc < 2)
        return 1;

    // MOV requires 3 args
    if(pArgs->argc < 3)
        return 1;

    // The base opcode
    uint8 opcode = 0x40; 

    for(int i = 1; i < 3; i++)
    {
        int argtype = GetArgType(pArgs, i);
        if(ARG_TYPE_REGISTER == argtype)
        {
            int iDest = ConvertReg(pArgs->argv[i][0]);
            opcode |= iDest<<( (i==1) ? 3 : 0 ); //shift it left 3 or none
        }
        else
        {
            opcode |= REG_NULL<<( (i==1) ? 3 : 0 ); // shift it left 3 or none
        }  
    }

    SaveToMemory(opcode);
    return 0;
}

int Assembler::HandleOpcodeMVI(AsmArgs* pArgs)
{
    if(pArgs->argc < 2)
        return 1;

    // MVI requires 3 args
    if(pArgs->argc < 3)
        return 1;

    // The base opcode
    uint8 opcode = 0x00; 

    //first arg is the destination
    int argtype = GetArgType(pArgs, 1);
    if(ARG_TYPE_REGISTER == argtype)
    {
        int iDest = ConvertReg(pArgs->argv[1][0]);
        opcode |= iDest<<3; //shift it left 3 or none
    }
    else
    if(ARG_TYPE_MEMORY == argtype)
    {
        opcode |= REG_NULL<<3; //shift it left 3 or none
    }


    opcode |= REG_NULL<<0; // MVI opcode always ends in REG_NULL or 110 in binary

    //second arg is the data in hex
    argtype = GetArgType(pArgs, 2);
    if(ARG_TYPE_HEX != argtype)
        return 1;

    char * pEnd;
    uint8 data = (uint8)strtoul(pArgs->argv[2], &pEnd, 16);

    // MVI
    SaveToMemory(opcode);

    // Data
    SaveToMemory(data);

    return 0;
}

int Assembler::HandleOpcodeADD(AsmArgs* pArgs)
{
    // ADD can only have 2 args
    if(pArgs->argc != 2)
    {
        return 1;
    }

    uint8 opcode = 0x80; // Set the base opcode
    int argtype = GetArgType(pArgs, 1);
    if(ARG_TYPE_REGISTER == argtype)
    {
        int iDest = ConvertReg(pArgs->argv[1][0]);
        opcode |= iDest<<0; //shift it left 0
    }
    else
    if(ARG_TYPE_MEMORY == argtype)
    {
        opcode |= REG_NULL<<0; //shift it left 0
    }
    else
    {
        return 1;
    }

    SaveToMemory(opcode);

    return 0;
}

int Assembler::HandleOpcodeCMP(AsmArgs* pArgs)
{
    // CMP can only have 2 args
    if(pArgs->argc != 2)
    {
        return 1;
    }

    uint8 opcode = 0xB8; // Set the base opcode
    int argtype = GetArgType(pArgs, 1);
    if(ARG_TYPE_REGISTER == argtype)
    {
        int iDest = ConvertReg(pArgs->argv[1][0]);
        opcode |= iDest<<0; //shift it left 0
    }
    else
    if(ARG_TYPE_MEMORY == argtype)
    {
        opcode |= REG_NULL<<0; //shift it left 0
    }
    else
    {
        return 1;
    }

    SaveToMemory(opcode);

    return 0;
}

int Assembler::HandleOpcodeRET(AsmArgs* pArgs)
{
    // ADD can only have 1 arg
    if(pArgs->argc > 1)
    {
        return 1;
    }

    uint8 opcode = 0xC9;
    SaveToMemory(opcode);

    return 0;
}

int  Assembler::ConvertReg(char cReg)
{    
    switch(cReg)
    {
    case 'A':
        return REG_A;
    case 'B':
        return REG_B;
    case 'C':
        return REG_C;
    case 'D':
        return REG_D;
    case 'E':
        return REG_E;
    case 'H':
        return REG_H;
    case 'L':
        return REG_L;
    default:
        {
            printf("Not a reg\n");
            return -1;
        }
        break;
    }

    return -1;
}

int  Assembler::GetArgType(AsmArgs* pArgs, int iIndex)
{
    if(iIndex > pArgs->argc)
        return ARG_TYPE_INVALID;

    int len = (int)strlen(pArgs->argv[iIndex]);

    if(len == 1) // Assume it's a register
    {
        return ARG_TYPE_REGISTER;
    }

    // If it's greater than 0 and not == 1, it MUST be a memory address
    if(len > 0)
    {
        //is it a valid hex?
        if (pArgs->argv[iIndex][len-1] == 'h')
        {
            return ARG_TYPE_HEX;
        }
        else
        if( strcmpi("Mem", pArgs->argv[iIndex])  == 0)
        {
            return ARG_TYPE_MEMORY;
        }

        return ARG_TYPE_INVALID;
    }

    return ARG_TYPE_INVALID;
}

void Assembler::SaveToMemory(uint8 data)
{
    m_Memory[m_MemoryCounter] = data;
    ++m_MemoryCounter;
}

void Assembler::Save(const char* pFilename)
{
    FILE* fp;
    fp = fopen(pFilename, "wb");
    if(fp)
    {
        fwrite(m_Memory, 1, m_MemoryCounter, fp);
        fclose(fp);
    }
}
