#include "stdafx.h"
#include "CaddyCPU.h"
#include "opcodedefs.h"

/*
#define FLAG_SIGN   (1 << 7)
#define FLAG_ZERO   (1 << 6)
#define FLAG_AC     (1 << 4)
#define FLAG_PARITY (1 << 2)
#define FLAG_CARRY  (1 << 0)
*/

CaddyCPU::CaddyCPU():
m_iProgramCounter(0),
m_iStackPointer(MEMORY_SIZE - 1)
{
    memset(&m_Reg, 0, sizeof(RegisterSet));
    memset(&m_Flags, 0, sizeof(FlagSet));
    memset(&m_Memory, 0, sizeof(uint8) * (MEMORY_SIZE));
}

CaddyCPU::~CaddyCPU()
{

}

RESULT CaddyCPU::Clear()
{
    // Reset all of our internals.
    m_iProgramCounter   = 0;
    m_iStackPointer     = MEMORY_SIZE - 1;
    memset(&m_Reg, 0, sizeof(RegisterSet));
    memset(&m_Flags, 0, sizeof(FlagSet));
    memset(&m_Memory, 0, sizeof(uint8) * (MEMORY_SIZE));

    return RESULT_OK;
}

// Cycle the CPU once
RESULT CaddyCPU::Tick()
{   
    // If we reach the end if memory, exit the mini program.
    if(m_iProgramCounter == MEMORY_SIZE)
    {
        return RESULT_DONE;
    }

    // Execute the instruction at the current instruction pointer
    RESULT res = ExecuteInstruction();

    return res;
}

RESULT CaddyCPU::SetMemoryAddress(uint16 iAddress, uint8 iValue)
{
    // Validate the input:
    if(iAddress > MEMORY_SIZE)
    {
        return RESULT_ERROR;
    }

    m_Memory[iAddress] = iValue;

    return RESULT_OK;
}

const uint8& CaddyCPU::GetMemoryAddress(uint16 iAddress) const
{
    // Validate the input:
    if(iAddress > MEMORY_SIZE)
    {
        return m_Memory[0];
    }

    return m_Memory[iAddress];
}

// We should remove this later maybe, since the mini program 
// should be the one manipulating the registers.
// It might be useful still however...
RESULT CaddyCPU::SetRegister(uint8 iRegister, uint8 iValue)
{
    if(iRegister > REG_A)
    {
        return RESULT_ERROR;
    }

    uint8 *pReg = (uint8*)(&m_Reg);
    pReg[iRegister] = iValue;


    return RESULT_OK;
}

const uint8& CaddyCPU::GetRegister(uint8 iReg)const
{
    uint8 *pReg = (uint8*)(&m_Reg);

    if(iReg > REG_A)
    {
        return pReg[REG_A];
    }

    return pReg[iReg];
}


// up to 0xFFFF
RESULT CaddyCPU::SetProgramCounter(uint16 iAddress)
{
    if(iAddress > MEMORY_SIZE)
    {
        return RESULT_ERROR;
    }

    m_iProgramCounter = iAddress;

    return RESULT_OK;
}

RESULT CaddyCPU::ExecuteInstruction()
{
    uint8 iInstruction = m_Memory[m_iProgramCounter];

    // Determine what this instruction is.

    // First check the mask for an MOV which is 01ffffffb
    // This is a speed optimization. Since MOV is the only opcode
    // to begin with 01ffffff, then we can mask it and single it out.
    // The other reason for this is since there are so many possible
    // "case:" values for MOV, this makes less of a readability nightmare.
    uint8 OpcodeMask = (iInstruction & 0xC0);

    //MOV
    if(OpcodeMask == 0x40)
    {
        return OPCODE_MOV();
    }

    switch(iInstruction)
    {
    //ADD
    case 0x80: // ADD B
    case 0x81: // ADD C
    case 0x82: // ADD D
    case 0x83: // ADD E
    case 0x84: // ADD H
    case 0x85: // ADD L
    case 0x87: // ADD A
    case 0x86: // ADD Memory (uses HL pair for address)
        {
            return OPCODE_ADD();
        }
        break;
    //MVI
    case 0x06: // MVI B
    case 0x0E: // MVI C
    case 0x16: // MVI D
    case 0x1E: // MVI E
    case 0x26: // MVI H
    case 0x2E: // MVI L
    case 0x3E: // MVI A
    case 0x36: // MVI Memory (uses HL pair for address)
        {
            return OPCODE_MVI();
        }
        break;
    case 0xB8: // CMP B
    case 0xB9: // CMP C
    case 0xBA: // CMP D
    case 0xBB: // CMP E
    case 0xBC: // CMP H
    case 0xBD: // CMP L
    case 0xBF: // CMP A
    case 0xBE: // CMP Mem (via HL pair)
        {
            return OPCODE_CMP();
        }
        break;
    //RET
    case 0xC9:
        {
            return OPCODE_RET();
        }
        break;
    default:
        {
        }
        break;
    }

    return RESULT_OK;
}

// Move immediate
RESULT CaddyCPU::OPCODE_MVI()
{
    uint8 iInstruction = m_Memory[m_iProgramCounter];

    //get the destination
    //need right shift it 3 when we're done
    uint8 iDest = (iInstruction & 0x38)>>3;

    if(iDest == REG_NULL)
    {
        // Move to memory address

        //Look in the HL register pair to find the memory address
        uint16 iMemoryAddress;
        iMemoryAddress = GetRegister(REG_H)<<(sizeof(uint8)*8);
        iMemoryAddress |= GetRegister(REG_L);

        // Get to the data
        ++m_iProgramCounter;
        m_Memory[iMemoryAddress] = m_Memory[m_iProgramCounter];

    }
    else
    {
        // Move to register

        // Get to the data
        ++m_iProgramCounter;
        //uint8 iData = m_Memory[m_iProgramCounter];

        uint8 *pReg = (uint8*)(&m_Reg);
        pReg[iDest] = m_Memory[m_iProgramCounter];
    }

    ++m_iProgramCounter;

    return RESULT_OK;
}

RESULT CaddyCPU::OPCODE_MOV()
{
    uint8 iInstruction = m_Memory[m_iProgramCounter];

    //assume this is a reg to reg copy for now
    
    //get the source
    uint8 iSource = (iInstruction & 0x07);

    //get the destination
    //need right shift it 3 when we're done
    uint8 iDest = (iInstruction & 0x38)>>3;

    //find out what type of reg MOV this is.
    // 1. Move from memory
    // 2. Move to memory
    // 3. Move to reg
    //
    if(REG_NULL == iSource)
    {
        //Move from memory

        //Look in the HL register pair to find the memory address
        uint16 iMemoryAddress;
        iMemoryAddress = GetRegister(REG_H)<<(sizeof(uint8)*8);
        iMemoryAddress |= GetRegister(REG_L);

        SetRegister(iDest,  m_Memory[iMemoryAddress]);       
    }
    else
    if(REG_NULL == iDest)
    {
        //Move to memory

        //Look in HL register pair to find the memory address
        uint16 iMemoryAddress;
        iMemoryAddress = GetRegister(REG_H)<<(sizeof(uint8)*8);
        iMemoryAddress |= GetRegister(REG_L);

        m_Memory[iMemoryAddress] = GetRegister(iSource);
    }
    else
    {
        //Move to register
        SetRegister(iDest, GetRegister(iSource));
    }

    ++m_iProgramCounter;

    return RESULT_OK;
}


RESULT CaddyCPU::OPCODE_ADD()
{
    uint8 iInstruction = m_Memory[m_iProgramCounter];

    //get the source
    uint8 iSource = (iInstruction & 0x07);

    if(REG_NULL == iSource)
    {
        // Add memory address in HL to Reg A

        //Look in HL register pair to find the memory address
        uint16 iMemoryAddress;
        iMemoryAddress = GetRegister(REG_H)<<(sizeof(uint8)*8);
        iMemoryAddress |= GetRegister(REG_L);

        uint8 *pReg = (uint8*)(&m_Reg);
        uint16 iCarry = pReg[REG_A]; //for carry calc
        pReg[REG_A] += m_Memory[iMemoryAddress];

        //TODO: set the carry bit, zero bit... etc...
        iCarry += m_Memory[iMemoryAddress];
        if(iCarry > pReg[REG_A])
        {
            m_Flags.carry = 1;
        }else
        {
            m_Flags.carry = 0;
        }

        if(pReg[REG_A] == 0)
        {
            m_Flags.zero = 1;
        }
        else
        {
            m_Flags.zero = 0;
        }


    }
    else
    {
        // Add source register to Reg A
        uint8 *pReg = (uint8*)(&m_Reg);
        pReg[REG_A] += pReg[iSource];

        //TODO: set the carry bit, zero bit... etc...
    }

    ++m_iProgramCounter;

    return RESULT_OK;
}

RESULT CaddyCPU::OPCODE_CMP()
{
    uint8 iInstruction = m_Memory[m_iProgramCounter];

    uint8 iSource = (iInstruction & 0x07);

    //TODO: refactor this code. It repeats code...

    // Is it a mem address?
    if(REG_NULL == iSource)
    {
        // Compare memory address in HL to Reg A

        //Look in HL register pair to find the memory address
        uint16 iMemoryAddress;
        iMemoryAddress = GetRegister(REG_H)<<(sizeof(uint8)*8);
        iMemoryAddress |= GetRegister(REG_L);

        uint8 *pReg = (uint8*)(&m_Reg);
        if(pReg[REG_A] == m_Memory[iMemoryAddress])
        {
            m_Flags.zero = 1;
        }else
        {
            m_Flags.zero = 0;
        }

        if(pReg[REG_A] < m_Memory[iMemoryAddress])
        {
            m_Flags.carry = 1;
        }
        else
        {
            m_Flags.carry = 0;
        }
    }
    else // It's a register
    {
        uint8 *pReg = (uint8*)(&m_Reg);
        if(pReg[REG_A] == pReg[iSource])
        {
            m_Flags.zero = 1;
        }else
        {
            m_Flags.zero = 0;
        }

        if(pReg[REG_A] < pReg[iSource])
        {
            m_Flags.carry = 1;
        }
        else
        {
            m_Flags.carry = 0;
        }
        
    }

    ++m_iProgramCounter;

    return RESULT_OK;
}

RESULT CaddyCPU::OPCODE_RET()
{
    if(m_iStackPointer == ((MEMORY_SIZE)-1) )
    {
        return RESULT_DONE;
    }

    // Set the Program Counter to the address stored on the stack.
    // This should only be called at the end of a CALL or
    // at the very end of a program.
    // It will produce undefined effects if otherwise used.

    // This is also little-endian code. 
    m_iProgramCounter = m_Memory[m_iStackPointer+1];
    m_iProgramCounter |= m_Memory[m_iStackPointer]<<8;
    m_iStackPointer -= 2;

    return RESULT_OK;
}